from collections.abc import Iterable
from dataclasses import dataclass
from typing import Final, Literal

from ast_grep_py import SgNode

from auto_typing_final.finder import (
    ImportsResult,
    find_all_definitions_in_functions,
    find_imports_of_identifier_in_scope,
    has_global_identifier_with_name,
)


@dataclass
class ImportConfig:
    value: str
    import_text: str
    import_identifier: str


ImportStyle = Literal["typing-final", "final"]
IMPORT_STYLES_TO_IMPORT_CONFIGS: dict[ImportStyle, ImportConfig] = {
    "typing-final": ImportConfig(value="typing.Final", import_text="import typing", import_identifier="typing"),
    "final": ImportConfig(value="Final", import_text="from typing import Final", import_identifier="Final"),
}


@dataclass
class EditableAssignmentWithoutAnnotation:
    node: SgNode
    left: str
    right: str


@dataclass
class EditableAssignmentWithAnnotation:
    node: SgNode
    left: str
    annotation: SgNode
    right: str


@dataclass
class OtherDefinition:
    node: SgNode


Definition = EditableAssignmentWithoutAnnotation | EditableAssignmentWithAnnotation | OtherDefinition


@dataclass
class AddFinal:
    node: Definition


@dataclass
class RemoveFinal:
    nodes: list[Definition]


Operation = AddFinal | RemoveFinal


def _make_definition_from_definition_node(node: SgNode) -> Definition:
    if node.kind() != "assignment":
        return OtherDefinition(node)

    match tuple((child.kind(), child) for child in node.children()):
        case (
            ("identifier", left),
            ("=", _),
            (right_kind, right),
        ) if right_kind != "assignment" and not ((parent := node.parent()) and parent.kind() == "assignment"):
            return EditableAssignmentWithoutAnnotation(node=node, left=left.text(), right=right.text())
        case (
            ("identifier", left),
            (":", _),
            ("type", annotation),
            ("=", _),
            (right_kind, right),
        ) if right_kind != "assignment" and not ((parent := node.parent()) and parent.kind() == "assignment"):
            return EditableAssignmentWithAnnotation(
                node=node, left=left.text(), annotation=annotation, right=right.text()
            )
        case _:
            return OtherDefinition(node)


def _make_operation_from_definitions_of_one_name(nodes: list[SgNode]) -> Operation:
    value_definitions: Final[list[Definition]] = []
    has_node_inside_loop = False

    for node in nodes:
        if any(ancestor.kind() in {"for_statement", "while_statement"} for ancestor in node.ancestors()):
            has_node_inside_loop = True
        value_definitions.append(_make_definition_from_definition_node(node))

    if has_node_inside_loop:
        return RemoveFinal(value_definitions)

    match value_definitions:
        case [definition]:
            return AddFinal(definition)
        case definitions:
            return RemoveFinal(definitions)


def _match_exact_identifier(node: SgNode, imports_result: ImportsResult, identifier_name: str) -> str | None:
    match tuple((inner_child.kind(), inner_child) for inner_child in node.children()):
        case (("identifier", first_identifier), (".", _), ("identifier", second_identifier)):
            for alias in imports_result.module_aliases:
                if alias == first_identifier.text() and second_identifier.text() == identifier_name:
                    return node.text()
    return None


def _strip_identifier_from_type_annotation(  # noqa: C901, PLR0911
    node: SgNode, imports_result: ImportsResult, identifier_name: str
) -> tuple[str, str] | None:
    type_node_children: Final = node.children()
    if len(type_node_children) != 1:
        return None
    inner_type_node: Final = type_node_children[0]
    kind: Final = inner_type_node.kind()

    if kind == "subscript":
        match tuple((child.kind(), child) for child in inner_type_node.children()):
            case (("attribute", attribute), ("[", _), *kinds_and_nodes, ("]", _)):
                if existing_final := _match_exact_identifier(attribute, imports_result, identifier_name):
                    return existing_final, "".join(node.text() for _, node in kinds_and_nodes)
    elif kind == "generic_type" and imports_result.has_from_import:
        match tuple((child.kind(), child) for child in inner_type_node.children()):
            case (("identifier", identifier), ("type_parameter", type_parameter)):
                if identifier.text() != identifier_name:
                    return None
                match tuple((inner_child.kind(), inner_child) for inner_child in type_parameter.children()):
                    case (("[", _), *kinds_and_nodes, ("]", _)):
                        return identifier_name, "".join(node.text() for _, node in kinds_and_nodes)
    elif kind == "identifier" and inner_type_node.text() == identifier_name:
        return identifier_name, ""
    elif kind == "attribute" and (
        existing_final := _match_exact_identifier(inner_type_node, imports_result, identifier_name)
    ):
        return existing_final, ""
    return None


def _make_changed_text_from_operation(  # noqa: C901
    operation: Operation, final_value: str, imports_result: ImportsResult, identifier_name: str
) -> Iterable[tuple[SgNode, str]]:
    match operation:
        case AddFinal(assignment):
            match assignment:
                case EditableAssignmentWithoutAnnotation(node, left, right):
                    yield node, f"{left}: {final_value} = {right}"
                case EditableAssignmentWithAnnotation(node, left, annotation, right):
                    match _strip_identifier_from_type_annotation(annotation, imports_result, identifier_name):
                        case None:
                            yield node, f"{left}: {final_value}[{annotation.text()}] = {right}"
                        case existing_final, "":
                            yield node, f"{left}: {existing_final} = {right}"
                        case existing_final, new_annotation:
                            yield node, f"{left}: {existing_final}[{new_annotation}] = {right}"

        case RemoveFinal(assignments):
            for assignment in assignments:
                match assignment:
                    case EditableAssignmentWithoutAnnotation(node, left, right):
                        yield node, f"{left} = {right}"
                    case EditableAssignmentWithAnnotation(node, left, annotation, right):
                        match _strip_identifier_from_type_annotation(annotation, imports_result, identifier_name):
                            case _, "":
                                yield node, f"{left} = {right}"
                            case _, new_annotation:
                                yield node, f"{left}: {new_annotation} = {right}"


@dataclass
class Edit:
    node: SgNode
    new_text: str


@dataclass
class Replacement:
    operation_type: type[Operation]
    edits: list[Edit]


@dataclass
class MakeReplacementsResult:
    replacements: list[Replacement]
    import_text: str | None


def make_replacements(root: SgNode, import_config: ImportConfig) -> MakeReplacementsResult:
    replacements: Final = []
    has_added_final = False
    imports_result: Final = find_imports_of_identifier_in_scope(root, module_name="typing", identifier_name="Final")

    for current_definitions in find_all_definitions_in_functions(root):
        operation = _make_operation_from_definitions_of_one_name(current_definitions)
        edits = [
            Edit(node=node, new_text=new_text)
            for node, new_text in _make_changed_text_from_operation(
                operation=operation,
                final_value=import_config.value,
                imports_result=imports_result,
                identifier_name="Final",
            )
            if node.text() != new_text
        ]

        if (operation_type := type(operation)) == AddFinal and edits:
            has_added_final = True

        replacements.append(Replacement(operation_type=operation_type, edits=edits))

    return MakeReplacementsResult(
        replacements=replacements,
        import_text=(
            import_config.import_text
            if has_added_final and not has_global_identifier_with_name(root=root, name=import_config.import_identifier)
            else None
        ),
    )



import argparse
import sys
from collections.abc import Iterable
from difflib import unified_diff
from pathlib import Path
from typing import Final, get_args

from ast_grep_py import SgRoot

from auto_typing_final.transform import IMPORT_STYLES_TO_IMPORT_CONFIGS, ImportConfig, ImportStyle, make_replacements


def transform_file_content(source: str, import_config: ImportConfig) -> str:
    root: Final = SgRoot(source, "python").root()
    result: Final = make_replacements(root, import_config)
    new_text: Final = root.commit_edits(
        [edit.node.replace(edit.new_text) for replacement in result.replacements for edit in replacement.edits]
    )
    return root.commit_edits([root.replace(f"{result.import_text}\n{new_text}")]) if result.import_text else new_text


def take_python_source_files(paths: Iterable[Path]) -> Iterable[Path]:
    for path in paths:
        if path.suffix in {".py", ".pyi"}:
            yield path


def find_source_files_from_one_path(path: Path) -> Iterable[Path]:
    if path.is_dir():
        for inner_path in path.iterdir():
            if inner_path.name.startswith("."):
                continue
            yield from take_python_source_files(find_source_files_from_one_path(inner_path))

    else:
        yield path


def find_all_source_files(paths: list[Path]) -> Iterable[Path]:
    for path in paths:
        yield from find_source_files_from_one_path(path)


def main() -> int:
    parser: Final = argparse.ArgumentParser()
    parser.add_argument("files", type=Path, nargs="*")
    parser.add_argument("--check", action="store_true")
    parser.add_argument("--import-style", type=str, choices=get_args(ImportStyle), default="typing-final")

    args: Final = parser.parse_args()
    import_config: Final = IMPORT_STYLES_TO_IMPORT_CONFIGS[args.import_style]

    has_changes = False

    for path in find_all_source_files(args.files):
        with path.open("r+") as file:
            source = file.read()
            transformed_content = transform_file_content(source=source, import_config=import_config)
            if source == transformed_content:
                continue

            has_changes = True

            if args.check:
                sys.stdout.writelines(
                    unified_diff(
                        source.splitlines(keepends=True),
                        transformed_content.splitlines(keepends=True),
                        fromfile=str(path),
                        tofile=str(path),
                    )
                )
            else:
                file.seek(0)
                file.write(transformed_content)
                file.truncate()

    return has_changes if args.check else 0



import pytest
from ast_grep_py import SgRoot

from auto_typing_final.finder import ImportsResult, find_imports_of_identifier_in_scope


@pytest.mark.parametrize(
    ("source", "result"),
    [
        ("import typing", ImportsResult(module_aliases={"typing"}, has_from_import=False)),
        ("import typing, platform", ImportsResult(module_aliases={"typing"}, has_from_import=False)),
        ("import typing, platform\nimport typing", ImportsResult(module_aliases={"typing"}, has_from_import=False)),
        ("from typing import Final", ImportsResult(module_aliases={"typing"}, has_from_import=True)),
        ("from typing import Final as Final", ImportsResult(module_aliases={"typing"}, has_from_import=False)),
        ("from typing import Final as F", ImportsResult(module_aliases={"typing"}, has_from_import=False)),
        ("import typing; from typing import Final", ImportsResult(module_aliases={"typing"}, has_from_import=True)),
        ("import typing as tp", ImportsResult(module_aliases={"tp", "typing"}, has_from_import=False)),
        (
            "import typing as tp\nimport typing as tt\nimport typing\n"
            "from typing import Final as F\nfrom typing import Final",
            ImportsResult(module_aliases={"tp", "tt", "typing"}, has_from_import=True),
        ),
        (
            "import typing as tp\nimport typing as tt",
            ImportsResult(module_aliases={"tp", "tt", "typing"}, has_from_import=False),
        ),
    ],
)
def test_get_global_imports(source: str, result: ImportsResult) -> None:
    assert (
        find_imports_of_identifier_in_scope(
            SgRoot(source, "python").root(), module_name="typing", identifier_name="Final"
        )
        == result
    )



import os
import sys
import uuid
from dataclasses import dataclass
from importlib.metadata import version
from pathlib import Path
from typing import Any, Final, TypedDict, cast
from urllib.parse import unquote_to_bytes

import attr
import cattrs
import lsprotocol.types as lsp
from ast_grep_py import SgRoot
from pygls.server import LanguageServer

from auto_typing_final.transform import (
    IMPORT_STYLES_TO_IMPORT_CONFIGS,
    AddFinal,
    Edit,
    ImportConfig,
    ImportStyle,
    make_replacements,
)


# From Python 3.13: https://github.com/python/cpython/blob/0790418a0406cc5419bfd9d718522a749542bbc8/Lib/pathlib/_local.py#L815
def path_from_uri(uri: str) -> Path | None:
    if not uri.startswith("file:"):
        return None
    path = uri[5:]
    if path[:3] == "///":
        # Remove empty authority
        path = path[2:]
    elif path[:12] == "//localhost/":
        # Remove 'localhost' authority
        path = path[11:]
    if path[:3] == "///" or (path[:1] == "/" and path[2:3] in ":|"):
        # Remove slash before DOS device/UNC path
        path = path[1:]
    if path[1:2] == "|":
        # Replace bar with colon in DOS drive
        path = path[:1] + ":" + path[2:]

    path_: Final = Path(os.fsdecode(unquote_to_bytes(path)))
    if not path_.is_absolute():
        return None
    return path_


ClientSettings = TypedDict("ClientSettings", {"import-style": ImportStyle})
FullClientSettings = TypedDict("FullClientSettings", {"auto-typing-final": ClientSettings})


@attr.define
class Fix:
    message: str
    text_edits: list[lsp.TextEdit | lsp.AnnotatedTextEdit]


def make_import_text_edit(import_text: str) -> lsp.TextEdit:
    return lsp.TextEdit(
        range=lsp.Range(start=lsp.Position(line=0, character=0), end=lsp.Position(line=0, character=0)),
        new_text=f"{import_text}\n",
    )


def make_text_edit(edit: Edit) -> lsp.TextEdit:
    node_range: Final = edit.node.range()
    return lsp.TextEdit(
        range=lsp.Range(
            start=lsp.Position(line=node_range.start.line, character=node_range.start.column),
            end=lsp.Position(line=node_range.end.line, character=node_range.end.column),
        ),
        new_text=edit.new_text,
    )


@dataclass
class Service:
    ls_name: str
    ignored_paths: list[Path]
    import_config: ImportConfig

    @staticmethod
    def try_from_settings(ls_name: str, settings: Any) -> "Service | None":  # noqa: ANN401
        try:
            validated_settings: Final = cattrs.structure(settings, FullClientSettings)
        except cattrs.BaseValidationError:
            return None

        executable_path: Final = Path(sys.executable)
        return Service(
            ls_name=ls_name,
            ignored_paths=[executable_path.parent.parent] if executable_path.parent.name == "bin" else [],
            import_config=IMPORT_STYLES_TO_IMPORT_CONFIGS[validated_settings["auto-typing-final"]["import-style"]],
        )

    def make_diagnostics(self, source: str) -> list[lsp.Diagnostic]:
        replacement_result: Final = make_replacements(
            root=SgRoot(source, "python").root(), import_config=self.import_config
        )
        result: Final = []

        for replacement in replacement_result.replacements:
            if replacement.operation_type == AddFinal:
                fix_message = f"{self.ls_name}: Add {self.import_config.value}"
                diagnostic_message = f"Missing {self.import_config.value}"
            else:
                fix_message = f"{self.ls_name}: Remove {self.import_config.value}"
                diagnostic_message = f"Unexpected {self.import_config.value}"

            fix = Fix(message=fix_message, text_edits=[make_text_edit(edit) for edit in replacement.edits])
            if replacement_result.import_text:
                fix.text_edits.append(make_import_text_edit(replacement_result.import_text))

            for applied_edit in replacement.edits:
                node_range = applied_edit.node.range()
                result.append(
                    lsp.Diagnostic(
                        range=lsp.Range(
                            start=lsp.Position(line=node_range.start.line, character=node_range.start.column),
                            end=lsp.Position(line=node_range.end.line, character=node_range.end.column),
                        ),
                        message=diagnostic_message,
                        severity=lsp.DiagnosticSeverity.Warning,
                        source=self.ls_name,
                        data=cattrs.unstructure(fix),
                    )
                )
        return result

    def make_fix_all_text_edits(self, source: str) -> list[lsp.TextEdit | lsp.AnnotatedTextEdit]:
        replacement_result: Final = make_replacements(
            root=SgRoot(source, "python").root(), import_config=self.import_config
        )
        result: Final[list[lsp.TextEdit | lsp.AnnotatedTextEdit]] = [
            make_text_edit(edit) for replacement in replacement_result.replacements for edit in replacement.edits
        ]
        if replacement_result.import_text:
            result.append(make_import_text_edit(replacement_result.import_text))
        return result

    def path_is_ignored(self, uri: str) -> bool:
        if path := path_from_uri(uri):
            return any(path.is_relative_to(ignored_path) for ignored_path in self.ignored_paths)
        return False


class CustomLanguageServer(LanguageServer):
    service: Service | None = None


LSP_SERVER = CustomLanguageServer(name="auto-typing-final", version=version("auto-typing-final"), max_workers=5)


@LSP_SERVER.feature(lsp.INITIALIZE)
def initialize(_: lsp.InitializeParams) -> None: ...


@LSP_SERVER.feature(lsp.INITIALIZED)
async def initialized(ls: CustomLanguageServer, _: lsp.InitializedParams) -> None:
    await ls.register_capability_async(
        params=lsp.RegistrationParams(
            registrations=[
                lsp.Registration(
                    id=str(uuid.uuid4()),
                    method=lsp.WORKSPACE_DID_CHANGE_CONFIGURATION,
                    register_options=lsp.DidChangeConfigurationRegistrationOptions(section=ls.name),
                ),
            ]
        )
    )


@LSP_SERVER.feature(lsp.WORKSPACE_DID_CHANGE_CONFIGURATION)
def workspace_did_change_configuration(ls: CustomLanguageServer, params: lsp.DidChangeConfigurationParams) -> None:
    ls.service = Service.try_from_settings(ls_name=ls.name, settings=params.settings) or ls.service
    if not ls.service:
        return
    for text_document in ls.workspace.text_documents.values():
        ls.publish_diagnostics(text_document.uri, diagnostics=ls.service.make_diagnostics(text_document.source))


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_SAVE)
@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
def did_open_did_save_did_change(
    ls: CustomLanguageServer,
    params: lsp.DidOpenTextDocumentParams | lsp.DidSaveTextDocumentParams | lsp.DidChangeTextDocumentParams,
) -> None:
    if not ls.service:
        return
    if ls.service.path_is_ignored(params.text_document.uri):
        return
    text_document: Final = ls.workspace.get_text_document(params.text_document.uri)
    ls.publish_diagnostics(text_document.uri, diagnostics=ls.service.make_diagnostics(text_document.source))


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_CLOSE)
def did_close(ls: CustomLanguageServer, params: lsp.DidCloseTextDocumentParams) -> None:
    ls.publish_diagnostics(params.text_document.uri, [])


@LSP_SERVER.feature(
    lsp.TEXT_DOCUMENT_CODE_ACTION,
    lsp.CodeActionOptions(
        code_action_kinds=[lsp.CodeActionKind.QuickFix, lsp.CodeActionKind.SourceFixAll], resolve_provider=True
    ),
)
def code_action(ls: CustomLanguageServer, params: lsp.CodeActionParams) -> list[lsp.CodeAction] | None:
    requested_kinds: Final = params.context.only or {lsp.CodeActionKind.QuickFix, lsp.CodeActionKind.SourceFixAll}
    actions: Final[list[lsp.CodeAction]] = []

    if lsp.CodeActionKind.QuickFix in requested_kinds:
        text_document: Final = ls.workspace.get_text_document(params.text_document.uri)
        our_diagnostics: Final = [
            diagnostic for diagnostic in params.context.diagnostics if diagnostic.source == ls.name
        ]

        for diagnostic in our_diagnostics:
            fix = cattrs.structure(diagnostic.data, Fix)
            actions.append(
                lsp.CodeAction(
                    title=fix.message,
                    kind=lsp.CodeActionKind.QuickFix,
                    edit=lsp.WorkspaceEdit(
                        document_changes=[
                            lsp.TextDocumentEdit(
                                text_document=lsp.OptionalVersionedTextDocumentIdentifier(
                                    uri=text_document.uri, version=text_document.version
                                ),
                                edits=fix.text_edits,
                            )
                        ]
                    ),
                    diagnostics=[diagnostic],
                )
            )

        if our_diagnostics:
            actions.append(
                lsp.CodeAction(
                    title=f"{ls.name}: Fix All",
                    kind=lsp.CodeActionKind.QuickFix,
                    data=params.text_document.uri,
                    edit=None,
                    diagnostics=params.context.diagnostics,
                )
            )

    if lsp.CodeActionKind.SourceFixAll in requested_kinds:
        actions.append(
            lsp.CodeAction(
                title=f"{ls.name}: Fix All",
                kind=lsp.CodeActionKind.SourceFixAll,
                data=params.text_document.uri,
                diagnostics=params.context.diagnostics,
            ),
        )

    return actions or None


@LSP_SERVER.feature(lsp.CODE_ACTION_RESOLVE)
def resolve_code_action(ls: CustomLanguageServer, params: lsp.CodeAction) -> lsp.CodeAction:
    if ls.service:
        text_document: Final = ls.workspace.get_text_document(cast(str, params.data))
        params.edit = lsp.WorkspaceEdit(
            document_changes=[
                lsp.TextDocumentEdit(
                    text_document=lsp.OptionalVersionedTextDocumentIdentifier(
                        uri=text_document.uri, version=text_document.version
                    ),
                    edits=ls.service.make_fix_all_text_edits(text_document.source),
                )
            ],
        )
    return params


def main() -> int:
    LSP_SERVER.start_io()
    return 0



from typing import Final

import pytest

from auto_typing_final.main import transform_file_content
from auto_typing_final.transform import IMPORT_STYLES_TO_IMPORT_CONFIGS, ImportConfig


@pytest.mark.parametrize("import_config", IMPORT_STYLES_TO_IMPORT_CONFIGS.values())
@pytest.mark.parametrize(
    ("before", "after"),
    [
        # Add annotation
        ("a: int", "a: int"),
        ("a = 1", "a: {} = 1"),
        ("a: {} = 1", "a: {} = 1"),
        ("a: int = 1", "a: {}[int] = 1"),
        ("a: typing.Annotated[int, 'hello'] = 1", "a: {}[typing.Annotated[int, 'hello']] = 1"),
        ("a: list[int] = 1", "a: {}[list[int]] = 1"),
        ("b = 1\na = 2\nb = 3", "b = 1\na: {} = 2\nb = 3"),
        ("b = 1\nb = 2\na = 3", "b = 1\nb = 2\na: {} = 3"),
        ("a = 1\nb = 2\nb = 3", "a: {} = 1\nb = 2\nb = 3"),
        ("a = 1\na = 2\nb: int", "a = 1\na = 2\nb: int"),
        ("a = 1\na: int", "a = 1\na: int"),
        ("a: int\na = 1", "a: int\na = 1"),
        ("a: {}\na = 1", "a: {}\na = 1"),
        ("a: int\na: int = 1", "a: int\na: int = 1"),
        ("a, b = 1, 2", "a, b = 1, 2"),
        ("(a, b) = 1, 2", "(a, b) = 1, 2"),
        ("(a, b) = t()", "(a, b) = t()"),
        ("[a, b] = t()", "[a, b] = t()"),
        ("[a] = t()", "[a] = t()"),
        ("a = b = 1", "a = b = 1"),
        ("a = b = c = 1", "a = b = c = 1"),
        ("a = (b := 1)", "a: {} = (b := 1)"),
        # Remove annotation
        ("a = 1\na: {}[int] = 2", "a = 1\na: int = 2"),
        ("a = 1\na: {} = 2", "a = 1\na = 2"),
        ("a: int = 1\na: {}[int] = 2", "a: int = 1\na: int = 2"),
        ("a: int = 1\na: {} = 2", "a: int = 1\na = 2"),
        ("a: {} = 1\na: {} = 2\na = 3\na: int = 4", "a = 1\na = 2\na = 3\na: int = 4"),
        ("a: {} = b = 1", "a: {} = b = 1"),
        # Both
        ("a = 1\nb = 2\nb: {}[int] = 3", "a: {} = 1\nb = 2\nb: int = 3"),
    ],
)
def test_variants(import_config: ImportConfig, before: str, after: str) -> None:
    source_function_content: Final = "\n".join(
        f"    {line.format(import_config.value)}" for line in before.splitlines()
    )
    source: Final = f"""
{import_config.import_text}

def foo():
{source_function_content}
"""

    after_function_content: Final = "\n".join(f"    {line.format(import_config.value)}" for line in after.splitlines())
    after_source: Final = f"""
{import_config.import_text}

def foo():
{after_function_content}
"""
    assert transform_file_content(source.strip(), import_config=import_config) == after_source.strip()


@pytest.mark.parametrize(
    "case",
    [
        """
a = 1
---
a = 1
""",
        """
def foo():
    a = 1
---
def foo():
    a: typing.Final = 1
""",
        """
a = 1

def foo():
    a = 2

    def bar():
        a = 3
---
a = 1

def foo():
    a: typing.Final = 2

    def bar():
        a: typing.Final = 3
""",
        """
a = 1

def foo():
    global a
    a = 2
---
a = 1

def foo():
    global a
    a = 2
""",
        """
def foo():
    from b import bar
    baz = 1
---
def foo():
    from b import bar
    baz: typing.Final = 1
""",
        """
def foo():
    from b import bar as baz
    bar = 1
    baz = 1
---
def foo():
    from b import bar as baz
    bar: typing.Final = 1
    baz = 1
""",
        """
def foo():
    from b import bar
    bar: typing.Final = 1
---
def foo():
    from b import bar
    bar = 1
""",
        """
def foo():
    import bar
    bar: typing.Final = 1
---
def foo():
    import bar
    bar = 1
""",
        """
def foo():
    import baz
    bar: typing.Final = 1
---
def foo():
    import baz
    bar: typing.Final = 1
""",
        """
def foo():
    from b import bar, baz
    bar = 1
    baz = 1
---
def foo():
    from b import bar, baz
    bar = 1
    baz = 1
""",
        """
def foo():
    from b import bar, baz as bazbaz
    bar = 1
    baz = 1
---
def foo():
    from b import bar, baz as bazbaz
    bar = 1
    baz: typing.Final = 1
""",
        """
def foo():
    # Dotted paths are not allowed, but tree-sitter-python grammar permits it
    from b import d.bar, bazbaz as baz
    bar = 1
    baz = 1
---
def foo():
    # Dotted paths are not allowed, but tree-sitter-python grammar permits it
    from b import d.bar, bazbaz as baz
    bar = 1
    baz = 1
""",
        """
def foo():
    from b import (bar, bazbaz)
    bar = 1
    baz = 1
---
def foo():
    from b import (bar, bazbaz)
    bar = 1
    baz: typing.Final = 1
""",
        """
def foo():
    a: typing.Final = 1
    a += 1
---
def foo():
    a = 1
    a += 1
""",
        """
def foo():
    a: typing.Final = 1
    a: int
---
def foo():
    a = 1
    a: int
""",
        """
def foo():
    a: typing.Final = 1
    a: typing.Final
---
def foo():
    a = 1
    a: typing.Final
""",
        """
def foo():
    a, b = 1
---
def foo():
    a, b = 1
""",
        """
def foo():
    a: typing.Final = 1
    b: typing.Final = 2
    a, b = 3
---
def foo():
    a = 1
    b = 2
    a, b = 3
""",
        """
def foo():
    a: typing.Final = 1
    b, c = 2
---
def foo():
    a: typing.Final = 1
    b, c = 2
""",
        """
def foo():
    a, b: typing.Final = 1
---
def foo():
    a, b: typing.Final = 1
""",
        """
def foo():
    a: typing.Final = 1
    (a, b) = 2
---
def foo():
    a = 1
    (a, b) = 2
""",
        """
def foo():
    a: typing.Final = 1
    (a, *other) = 2
---
def foo():
    a = 1
    (a, *other) = 2
""",
        """
def foo():
    def a(): ...
    a: typing.Final = 1
---
def foo():
    def a(): ...
    a = 1
""",
        """
def foo():
    class a: ...
    a: typing.Final = 1
---
def foo():
    class a: ...
    a = 1
""",
        """
def foo():
    a: typing.Final = 1
    if a := 1: ...
---
def foo():
    a = 1
    if a := 1: ...
""",
        """
def foo():
    while True:
        a = 1
---
def foo():
    while True:
        a = 1
""",
        """
def foo():
    while True:
        a: typing.Final = 1
---
def foo():
    while True:
        a = 1
""",
        """
def foo():
    for _ in ...:
        a: typing.Final = 1
---
def foo():
    for _ in ...:
        a = 1
""",
        """
def foo():
    for _ in ...:
        def foo():
            a: typing.Final = 1
---
def foo():
    for _ in ...:
        def foo():
            a = 1
""",
        """
def foo():
    a: typing.Final = 1
    b: typing.Final = 2

    for _ in ...:
        a: typing.Final = 1
---
def foo():
    a = 1
    b: typing.Final = 2

    for _ in ...:
        a = 1
""",
        """
def foo():
    for _ in ...:
        a = 1
---
def foo():
    for _ in ...:
        a = 1
""",
        """
def foo():
    a: typing.Final = 1
    for a in ...: ...
---
def foo():
    a = 1
    for a in ...: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case ...: ...
---
def foo():
    a: typing.Final = 1

    match ...:
        case ...: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case [] as a: ...
---
def foo():
    a = 1

    match ...:
        case [] as a: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case {"hello": a, **b}: ...
---
def foo():
    a = 1

    match ...:
        case {"hello": a, **b}: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case {**a}: ...
---
def foo():
    a = 1

    match ...:
        case {**a}: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case A(b=a) | B(b=a): ...
---
def foo():
    a = 1

    match ...:
        case A(b=a) | B(b=a): ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case [b, *a]: ...
---
def foo():
    a = 1

    match ...:
        case [b, *a]: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case [a]: ...
---
def foo():
    a = 1

    match ...:
        case [a]: ...
""",
        """
def foo():
    a: typing.Final = 1

    match ...:
        case (a,): ...
---
def foo():
    a = 1

    match ...:
        case (a,): ...
""",
        """
def foo():
    a: typing.Final = 1
    nonlocal a
---
def foo():
    a = 1
    nonlocal a
""",
        """
def foo():
    a = 1
    nonlocal a
---
def foo():
    a = 1
    nonlocal a
""",
        """
def foo():
    a: typing.Final = 1
    global b
---
def foo():
    a: typing.Final = 1
    global b
""",
        """
def foo():
    a: typing.Final = 1
    global a
---
def foo():
    a = 1
    global a
""",
        """
def foo():
    a: typing.Final = 1
    b: typing.Final = 2
    c: typing.Final = 3

    def bar():
        nonlocal a
        b: typing.Final = 4
        c: typing.Final = 5

        class C:
            a = 6
            c = 7

            def baz():
                nonlocal a, b
                b: typing.Final = 8
                c: typing.Final = 9
---
def foo():
    a = 1
    b: typing.Final = 2
    c: typing.Final = 3

    def bar():
        nonlocal a
        b = 4
        c: typing.Final = 5

        class C:
            a = 6
            c = 7

            def baz():
                nonlocal a, b
                b = 8
                c: typing.Final = 9
""",
        """
def foo():
    foo = 1
---
def foo():
    foo: typing.Final = 1
""",
        """
def foo(a, b: int, c=1, d: int = 2):
    a: typing.Final = 1
    b: typing.Final = 2
    c: typing.Final = 3
    d: typing.Final = 4
    e: typing.Final = 5
---
def foo(a, b: int, c=1, d: int = 2):
    a = 1
    b = 2
    c = 3
    d = 4
    e: typing.Final = 5
""",
        """
def foo(self):
    self.me = 1
---
def foo(self):
    self.me = 1
""",
        """
a.b = 1
---
a.b = 1
""",
    ],
)
def test_transform_file_content(case: str) -> None:
    import_config: Final = IMPORT_STYLES_TO_IMPORT_CONFIGS["typing-final"]
    before, _, after = case.partition("---")
    assert (
        transform_file_content(f"{import_config.import_text}\n" + before.strip(), import_config=import_config)
        == f"{import_config.import_text}\n" + after.strip()
    )


@pytest.mark.parametrize(
    "case",
    [
        """
import typing

def f():
    a = 1
---
import typing

def f():
    a: typing.Final = 1
""",
        """
import typing

def f():
    a: typing.Final = 1
---
import typing

def f():
    a: typing.Final = 1
""",
        """
def f():
    a = 1
---
import typing
def f():
    a: typing.Final = 1
""",
        """
def f():
    a: typing.Final = 1
    a = 2
---
def f():
    a = 1
    a = 2
""",
        """
typing = 1

def f():
    a = 1
---
typing = 1

def f():
    a: typing.Final = 1
""",
    ],
)
def test_add_import(case: str) -> None:
    before, _, after = case.partition("---")
    import_config: Final = IMPORT_STYLES_TO_IMPORT_CONFIGS["typing-final"]
    assert transform_file_content(before.strip(), import_config=import_config) == after.strip()


@pytest.mark.parametrize(
    ("case", "import_config"),
    [
        (
            """
from typing import Final

def f():
    a: Final[int] = 1
---
from typing import Final

def f():
    a: Final[int] = 1
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["typing-final"],
        ),
        (
            """
def f():
    a: Final = 1
    b: typing.Final = 2
    c: Final = 3
---
def f():
    a: Final = 1
    b: typing.Final = 2
    c: Final = 3
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["typing-final"],
        ),
        (
            """
import typing

def f():
    a: Final = 1
    b: typing.Final = 2
    c: Final = 3
---
import typing

def f():
    a: Final = 1
    b: typing.Final = 2
    c: Final = 3
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["final"],
        ),
        (
            """
import typing as tp

def f():
    a: tp.Final = 1
---
import typing as tp

def f():
    a: tp.Final = 1
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["typing-final"],
        ),
        (
            """
import typing as tp

def f():
    a: tp.Final = 1
---
import typing as tp

def f():
    a: tp.Final = 1
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["final"],
        ),
        (
            """
import typing as tp

def f():
    a: tp.Final = 1
    a: tp.Final = 2
    b = 1
---
import typing
import typing as tp

def f():
    a = 1
    a = 2
    b: typing.Final = 1
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["typing-final"],
        ),
        (
            """
import typing as tp

def f():
    a: tp.Final = 1
    a: tp.Final = 2
    b = 1
---
from typing import Final
import typing as tp

def f():
    a = 1
    a = 2
    b: Final = 1
""",
            IMPORT_STYLES_TO_IMPORT_CONFIGS["final"],
        ),
    ],
)
def test_different_styles(case: str, import_config: ImportConfig) -> None:
    before, _, after = case.partition("---")
    assert transform_file_content(before.strip(), import_config=import_config) == after.strip()



from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Final

from ast_grep_py import Config, SgNode

# https://github.com/tree-sitter/tree-sitter-python/blob/71778c2a472ed00a64abf4219544edbf8e4b86d7/grammar.js
DEFINITION_RULE: Config = {
    "rule": {
        "any": [
            {"kind": "assignment"},
            {"kind": "augmented_assignment"},
            {"kind": "named_expression"},
            {"kind": "function_definition"},
            {"kind": "global_statement"},
            {"kind": "nonlocal_statement"},
            {"kind": "class_definition"},
            {"kind": "import_from_statement"},
            {"kind": "import_statement"},
            {"kind": "as_pattern"},
            {"kind": "keyword_pattern"},
            {"kind": "splat_pattern"},
            {"kind": "dict_pattern"},
            {"kind": "list_pattern"},
            {"kind": "tuple_pattern"},
            {"kind": "for_statement"},
        ]
    }
}


def _get_last_child_of_type(node: SgNode, type_: str) -> SgNode | None:
    return last_child if (children := node.children()) and (last_child := children[-1]).kind() == type_ else None


def _find_identifiers_in_children(node: SgNode) -> Iterable[SgNode]:
    for child in node.children():
        if child.kind() == "identifier":
            yield child


def _is_inside_inner_function(root: SgNode, node: SgNode) -> bool:
    for ancestor in node.ancestors():
        if ancestor.kind() == "function_definition":
            return ancestor != root
    return False


def _is_inside_inner_function_or_class(root: SgNode, node: SgNode) -> bool:
    for ancestor in node.ancestors():
        if ancestor.kind() in {"function_definition", "class_definition"}:
            return ancestor != root
    return False


def _find_identifiers_in_import_statement(node: SgNode) -> Iterable[SgNode]:
    match tuple((child.kind(), child) for child in node.children()):
        case (("from", _), _, ("import", _), *name_nodes) | (("import", _), *name_nodes):
            for kind, name_node in name_nodes:
                match kind:
                    case "dotted_name":
                        if identifier := _get_last_child_of_type(name_node, "identifier"):
                            yield identifier
                    case "aliased_import":
                        if alias := name_node.field("alias"):
                            yield alias


def _find_identifiers_made_by_node(node: SgNode) -> Iterable[SgNode]:  # noqa: C901, PLR0912
    match node.kind():
        case "assignment" | "augmented_assignment":
            if not (left := node.field("left")):
                return
            match left.kind():
                case "pattern_list" | "tuple_pattern":
                    yield from _find_identifiers_in_children(left)
                case "identifier":
                    yield left
        case "named_expression":
            if name := node.field("name"):
                yield name
        case "class_definition":
            if name := node.field("name"):
                yield name
            for function in node.find_all(kind="function_definition"):
                for nonlocal_statement in node.find_all(kind="nonlocal_statement"):
                    if _is_inside_inner_function(root=function, node=nonlocal_statement):
                        continue
                    yield from _find_identifiers_in_children(nonlocal_statement)
        case "function_definition":
            if name := node.field("name"):
                yield name
            for nonlocal_statement in node.find_all(kind="nonlocal_statement"):
                if _is_inside_inner_function(root=node, node=nonlocal_statement):
                    continue
                yield from _find_identifiers_in_children(nonlocal_statement)
        case "import_from_statement" | "import_statement":
            yield from _find_identifiers_in_import_statement(node)
        case "as_pattern":
            match tuple((child.kind(), child) for child in node.children()):
                case (
                    (("identifier", _), ("as", _), ("as_pattern_target", alias))
                    | (("case_pattern", _), ("as", _), ("identifier", alias))
                ):
                    yield alias
        case "keyword_pattern":
            match tuple((child.kind(), child) for child in node.children()):
                case (("identifier", _), ("=", _), ("dotted_name", alias)):
                    if identifier := _get_last_child_of_type(alias, "identifier"):
                        yield identifier
        case "list_pattern" | "tuple_pattern":
            for child in node.children():
                if (
                    child.kind() == "case_pattern"
                    and (dotted_name := _get_last_child_of_type(child, "dotted_name"))
                    and (identifier := _get_last_child_of_type(dotted_name, "identifier"))
                ):
                    yield identifier
        case "splat_pattern" | "global_statement" | "nonlocal_statement":
            yield from _find_identifiers_in_children(node)
        case "dict_pattern":
            for child in node.children():
                if (
                    child.kind() == "case_pattern"
                    and (previous_child := child.prev())
                    and previous_child.kind() == ":"
                    and (dotted_name := _get_last_child_of_type(child, "dotted_name"))
                    and (identifier := _get_last_child_of_type(dotted_name, "identifier"))
                ):
                    yield identifier
        case "for_statement":
            if left := node.field("left"):
                yield from left.find_all(kind="identifier")


def _find_identifiers_in_current_scope(node: SgNode) -> Iterable[tuple[SgNode, SgNode]]:
    for child in node.find_all(DEFINITION_RULE):
        if _is_inside_inner_function_or_class(node, child) or child == node:
            continue
        for identifier in _find_identifiers_made_by_node(child):
            yield identifier, child


def _find_identifiers_in_function_parameter(node: SgNode) -> Iterable[SgNode]:
    match node.kind():
        case "default_parameter" | "typed_default_parameter":
            if name := node.field("name"):
                yield name
        case "identifier":
            yield node
        case _:
            yield from _find_identifiers_in_children(node)


def find_all_definitions_in_functions(root: SgNode) -> Iterable[list[SgNode]]:
    for function in root.find_all(kind="function_definition"):
        definition_map = defaultdict(list)

        if parameters := function.field("parameters"):
            for parameter in parameters.children():
                for identifier in _find_identifiers_in_function_parameter(parameter):
                    definition_map[identifier.text()].append(parameter)

        for identifier, node in _find_identifiers_in_current_scope(function):
            definition_map[identifier.text()].append(node)

        yield from definition_map.values()


def has_global_identifier_with_name(root: SgNode, name: str) -> bool:
    return name in {identifier.text() for identifier, _ in _find_identifiers_in_current_scope(root)}


@dataclass
class ImportsResult:
    module_aliases: set[str]
    has_from_import: bool


def find_imports_of_identifier_in_scope(root: SgNode, module_name: str, identifier_name: str) -> ImportsResult:  # noqa: C901
    result: Final = ImportsResult(module_aliases={module_name}, has_from_import=False)

    for node in root.find_all(any=[{"kind": "import_statement"}, {"kind": "import_from_statement"}]):
        if _is_inside_inner_function_or_class(root, node) or node == root:
            continue

        match tuple((child.kind(), child) for child in node.children()):
            case (("from", _), (_, module), ("import", _), *name_nodes):
                for kind, name_node in name_nodes:
                    if (
                        kind == "dotted_name"
                        and (identifier := _get_last_child_of_type(name_node, "identifier"))
                        and module.text() == module_name
                        and identifier.text() == identifier_name
                    ):
                        result.has_from_import = True
            case (("import", _), *name_nodes):
                for kind, name_node in name_nodes:
                    match kind:
                        case "dotted_name":
                            if (identifier := _get_last_child_of_type(name_node, "identifier")) and (
                                identifier_text := identifier.text()
                            ) == module_name:
                                result.module_aliases.add(identifier_text)
                        case "aliased_import":
                            if (
                                (name := name_node.field("name"))
                                and (name.text() == module_name)
                                and (alias := name_node.field("alias"))
                            ):
                                result.module_aliases.add(alias.text())
    return result
